(********************************************************************)
(* SessML - Runtime                                                 *)
(*                                                                  *)
(* data.ml                                                       *)
(********************************************************************)
(* Time-stamp: <Pierre-Malo Denielou - 2010> *)

open Global

(**********************************************)
(* String operations and message construction *)
(**********************************************)

let int_to_4string i =
  (if i<10 then "000" else 
    (if i<100 then "00" else
	(if i<1000 then "0" else "")))
    ^(string_of_int i)

let concatold x y = 
  let ret = 
    (int_to_4string (String.length x + String.length y))
    ^(int_to_4string (String.length x))
    ^x^y in
  debug_concat (Printf.sprintf "Concatenating %d + %d = %d"
                  (String.length x) (String.length y) (String.length ret));
  ret

let iconcatold z = 
  debug_concat (Printf.sprintf "Attempting decat %d: " (String.length z));
  let total = int_of_string (String.sub z 0 4) in
  let shift = int_of_string (String.sub z 4 4) in
  debug_concat (Printf.sprintf "total: %d shift: %d" total shift);
  let x1 = String.sub z 8 shift in
  let x2 = String.sub z (shift+8) (total-shift) in 
  debug_concat (Printf.sprintf "success %d %d!" (String.length x1) (String.length x2));
  (x1,x2)

let concat x y = 
  let n = String.length x in
  let m = String.length y in
  let b = Buffer.create (n+m+8) in
  Buffer.add_string b (int_to_4string (n+m));
  Buffer.add_string b (int_to_4string n);
  Buffer.add_string b x;
  Buffer.add_string b y;
  Buffer.contents b

let iconcat z = 
  debug_concat (Printf.sprintf "Attempting decat %d: " (String.length z));
  let total = int_of_string (String.sub z 0 4) in
  let shift = int_of_string (String.sub z 4 4) in
  debug_concat (Printf.sprintf "total: %d shift: %d" total shift);
  let x1 = String.sub z 8 shift in
  let x2 = String.sub z (shift+8) (total-shift) in 
  debug_concat (Printf.sprintf "success %d %d!" (String.length x1) (String.length x2));
  (x1,x2)
    
let concat3 x y z = concat x (concat y z)
let iconcat3 e = (let x,yz = iconcat e in let y,z = iconcat yz in x,y,z)

let addtuple b n l =
  let rec aux = function
    | [] -> ()
    | a::q -> (let n = String.length a in 
               Buffer.add_string b (int_to_4string n); 
               Buffer.add_string b a;
               aux q)
  in
  aux l

let concattuple n l =
  let b = Buffer.create (n*8) in
  addtuple b n l;
  Buffer.contents b
  
let iconcattuple ?(off=0) n z =
  let rec aux i p =
    if i<n then
      let n = int_of_string (String.sub z p 4) in
      let s = String.sub z (p+4) n in
      s::(aux (i+1) (p+4+n))
    else [] in
  aux 0 off
  

let concatlist l = 
  let n = List.length l in
  let b = Buffer.create (n*8+4) in
  Buffer.add_string b (int_to_4string n);
  addtuple b n l;
  Buffer.contents b

let iconcatlist z =
  let n = int_of_string (String.sub z 0 4) in
  debug_concat (Printf.sprintf "List of size %d" n);
  iconcattuple ~off:4 n z

type str = string

let cS (x:string) : str = x
let iS (x:str) : string = x


(* Building message *)

type session_message=
    {session:string;
     mnb:string;
     srole:string;
     sid:string;
     rrole:string;
     rid:string;
     lbl:string;
     lv:string list;
     body:string} 

let buildmsg session mnb srole sid rrole rid label lv content =
  let option = concatlist [mnb;sid;rid;(concatlist lv)] in
  concattuple 6 [session;srole;rrole;label;content;option]
(*  concattuple 9 [session;mnb;srole;sid;rrole;rid;label;(concatlist lv);content]*)

let readmsg m =
  match iconcattuple 6 m with
    | [session;srole;rrole;label;body;option] -> begin
        match iconcatlist option with
            [mnb;sid;rid;slv] -> 
              {session=session;lbl=label;lv=(iconcatlist slv);mnb=mnb;
               srole=srole;sid=sid;rrole=rrole;rid=rid;body=body}
          | _ -> 
              {session=session;lbl=label;lv=[];mnb="";
               srole=srole;sid="";rrole=rrole;rid="";body=body}
      end
    | _ -> assert false
(*
  match iconcattuple 9 m with
      [session;mnb;srole;sid;rrole;rid;label;lv;body] ->
        {session=session;lbl=label;lv=(iconcatlist lv);mnb=mnb;srole=srole;sid=sid;rrole=rrole;rid=rid;body=body}
    | _ -> assert false
*)

let sprintf_msg msg =
  Printf.sprintf "Received a message %s (nb %s) from %s:%s to %s:%s"
    msg.lbl msg.mnb msg.srole msg.sid msg.rrole msg.rid

