(********************************************************************)
(* SessML - Runtime                                                 *)
(*                                                                  *)
(* prins.ml                                                       *)
(********************************************************************)
(* Time-stamp: <Pierre-Malo Denielou - 2010> *)

open Unix
open Global
open Data


(**********************)
(* Principal database *)
(**********************)

type pr = {id:string; ip:string; port:int}

let addrDb : (string * (string * int)) list ref = ref []

let register (prin:pr) =
  debug ("Registering "^prin.id);
  if List.mem_assoc prin.id !addrDb
  then (if List.assoc prin.id !addrDb = (prin.ip,prin.port)
    then debug (prin.id^" already known") 
    else failwith ("Principal identity confusion in database"))
  else
    addrDb := (prin.id,(prin.ip,prin.port))::!addrDb

let reg_read idl =
  debug ("Updating database") ;
  let ls = iconcatlist idl in
  let l = List.map (fun s ->
    let (r,net) = iconcat s in
    let ip,pos = iconcat net in
    let po = int_of_string pos in
    {id=r;ip=ip;port=po}) ls in
  List.iter register l;
  debug ("Database updated") ;
  ()


let reg_write reg =
  let rec aux = function
    | [] -> []
    | (role,l)::q -> l@(aux q)
  in
  let l = remove_duplicates (aux reg) in
  let ls = List.map 
    (function id -> 
      let ip,po = List.assoc id !addrDb in
      concat id (concat ip (string_of_int po))
    ) l in
  concatlist ls

(* Local registry *)

let lreg_get reg role sub lv =
  let lv = List.map (alias sub) lv in
  let lp = List.assoc role reg in
  substract lv lp

let lsub_get reg =
  let rec aux = function
      [] -> []
    | (role,[p])::q -> (role,p)::(aux q)
    | _::q -> aux q
  in
  aux reg
  
                
