(********************************************************************)
(* SessML - Runtime                                                 *)
(*                                                                  *)
(* tcp.ml                                                           *)
(********************************************************************)
(* Time-stamp: <Deniélou Malo - 2010> *)

open Unix
open Global
open Data
open Prins

(*****************)
(* Communication *)
(*****************)

let out_channel = ref []
let  in_descr = ref []

let add_chan addr sock =
  let outc = Unix.out_channel_of_descr sock in
  out_channel:= (addr,outc):: !out_channel;
  in_descr := (sock):: !in_descr

let get_out_channel addr =
  if List.mem_assoc addr !out_channel
  then List.assoc addr !out_channel
  else begin
    let sock = Unix.socket Unix.PF_INET Unix.SOCK_STREAM 0 in
    Unix.connect sock addr;
    add_chan addr sock;
    List.assoc addr !out_channel
  end
  
let listen = ref true

let create_tcp_server ?(max=None) ?(condition=None) ~address ~port () =
  let socket = Unix.socket Unix.PF_INET Unix.SOCK_STREAM 0 in
  let accept_thread () =
    let count = ref 0 in
    while !listen && (match max with None -> true | Some max -> !count < max) do
      let (sock,addr) = Unix.accept socket in
      add_chan addr sock;
      debug_com (Printf.sprintf "Connection accepted! ");
      incr count;
      match condition with
          None -> ()
        | Some c -> Condition.signal c
    done;
    ()
  in
  try 
    debug_com (Printf.sprintf "Listening on port %d " port);
    Unix.bind socket (Unix.ADDR_INET (Unix.inet_addr_of_string address,port));
    Unix.listen socket 5;
    Thread.create accept_thread ()
  with
    | Unix_error (e,_,_) -> (
        debug (Printf.sprintf "Unix error: %s" (error_message e));
        assert false)

  
let bind prin =
  let (ip,localport) = List.assoc prin (!addrDb) in
  create_tcp_server ~address:ip ~port:localport ()


let close () = 
  List.iter (fun x -> Unix.shutdown x Unix.SHUTDOWN_ALL) !in_descr;
  in_descr := [];
  out_channel:=[];
  ()

let send ?(verb=true) (dest:string) (text:str): unit = 
  let (url,destport) = List.assoc dest !addrDb in
  let inet = Unix.inet_addr_of_string url in
  let addr = Unix.ADDR_INET (inet,destport) in
  let outc = get_out_channel addr in
  output_string outc text;
  output_char outc '\n';
  flush outc;
  debug_com (Printf.sprintf "Message sent!" ) ;
  (* let () = Hashtbl.add acktable mnb (dest,txt) in *)
  if verb then debug_com ("sent to "^dest^":\n"^text) ;
  ()

let psend ?(verb=true) acktable mnb (dest:string) (text:str): unit = 
  send ~verb dest text;
  ()


let messages = Queue.create ()

let precv () : str option =
  let rec aux = function
      [] -> None
    | s::rest -> 
        let text = String.create 2056 in
        let len = recv s text 0 2056 [] in
        if len=0 then aux rest
        else begin
          debug_com ("A message is received!") ;
          let text = String.sub text 0 len in
          debug_com (Printf.sprintf "Content is: %S" (String.escaped text)) ;
          let ls = split text in
          (match ls with
               [] -> (debug_com ("Empty message ?!") ; aux rest)
             | m::q -> (List.iter (fun x -> Queue.push x messages) q; Some m))
        end 
  in
  if Queue.is_empty messages
  then begin
    try 
      let lr,_,le= Unix.select !in_descr [] !in_descr 0.1 in
      if le <> [] then begin 
        debug_com ("A network error is found!") ;
        List.iter Unix.close le;
        in_descr := substract_eq le !in_descr;
        None
      end
      else 
        aux lr
    with
      | Unix_error (ECONNRESET,_,_) -> (
          debug (Printf.sprintf "Connection reset error");
          assert false)
      | Unix_error (e,_,_) -> (
          debug (Printf.sprintf "Unix error: %s" (error_message e));
          assert false)
  end
  else Some (Queue.pop messages)

let sendAck session srole sid rrole rid mnb =
  ()
