(********************************************************************)
(* SessML - Runtime                                                 *)
(*                                                                  *)
(* udp.ml                                                       *)
(********************************************************************)
(* Time-stamp: <Deniélou Malo - 2011> *)

open Unix
open Global
open Data
open Prins


(*****************)
(* Communication *)
(*****************)

let udp = ref (Unix.stderr)

let local = ref ("")

let bind ?(interactive=false) prin =
  let (_,localport) = List.assoc prin (!addrDb) in
  debug_com ("Listening on port "^(string_of_int localport));
  udp := Unix.socket Unix.PF_INET Unix.SOCK_DGRAM 0;
  Unix.setsockopt !udp SO_REUSEADDR true;
  Unix.setsockopt_float !udp SO_RCVTIMEO 0.1;
  Unix.bind !udp (Unix.ADDR_INET (Unix.inet_addr_loopback, localport))

let close () = 
  Unix.close !udp

let netlock = Mutex.create ()

(* Low-level communication *)
let send ?(verb=true) (dest:string) (text:str): unit = 
  Mutex.lock netlock;
  let (url,destport) = List.assoc dest !addrDb in
  let text = iS text in
  let udpSend = socket PF_INET SOCK_DGRAM 0 in
  let addr = inet_addr_of_string url in
  let _ = sendto udpSend text 0 (String.length text) [] (ADDR_INET(addr,destport)) in 
  let _ = Unix.close udpSend in
  debug_com (Printf.sprintf "Message sent!" ) ;
  if verb then debug_com ("sent to "^dest^":\n"^text) ;
  Mutex.unlock netlock;
  ()


let psend ?(exchange="") ?(verb=true) acktable mnb (dest:string) (text:str)=
  send ~verb dest text;
  let () = Hashtbl.add acktable mnb (dest,text) in
  debug_ack (Printf.sprintf "Expecting an Ack for %s from %s" mnb dest) ;
  ()


let sendAck session srole sid rrole rid mnb =
  let txt = buildmsg session mnb srole sid rrole rid "ack" [] "" in
  debug_ack (Printf.sprintf "Sending acknowlegement message %s from %s:%s to %s:%s" mnb srole
           sid rrole rid) ;
  try send rid txt ; true 
  with Not_found -> (
    debug_ack ("Sender not in database");
    Mutex.unlock netlock;
    false)


let precv () : str option =
  Mutex.lock netlock;
  let text = String.create 2056 in
  (*   let (len,_) = recvfrom udp text 0 1024 [] in *)
  try 
    let len = recv !udp text 0 2056 [] in
    debug_com ("A message is received!") ;
    debug_com ("Content is:\n"^text) ;
    Mutex.unlock netlock;
    Some (cS (String.sub text 0 len))
  with
      Unix_error (EAGAIN,_,_) -> (
        Mutex.unlock netlock;
        None)
    | Unix_error (e,_,_) -> (
        debug (Printf.sprintf "Unix error: %s" (error_message e));
        Mutex.unlock netlock;
        assert false)

let init p = ()
let init_session sessionid lr = ()
let send_Basic_Consume ?(queue="") ?(ctag="") () = ()
