(********************************************************************)
(* SessML - Implementation                                          *)
(*                                                                  *)
(* wf.ml                                                            *)
(********************************************************************)
(* $Time-stamp: <Deniélou Malo - 2010>$ *)

open Common
open Syntax

let debug_wf = gen_debug debug_wf "Well-formedness"

(* Conversion to global type *)
let global_conversion (rl:roles_list) (g:as_global) =
  let rk = List.map (function (r,k,_)-> (r,k)) rl in
  
  let rec muelimination (bv: p list) (next:node option) (gend:global_type option) = function
    | GASRec(_,g) -> assert false
    | GASOne(_,g) -> assert false
    | GASInv(_,g) -> assert false
    | GASEnd -> 
        let n = match next with None -> create_node_name () | Some k -> k in
        (match gend with None -> GEnd n | Some g -> g)
    | GASGoto (_,g) -> 
        GGoto (add_ident g)
    | GASMu (_,g,r) ->
        let n = add_ident g in
        muelimination bv (Some n) gend r
    | GASMsg (_,(s,r,ml)) -> 
        let n = match next with None -> create_node_name () | Some k -> k in
        let ss,bbv = try (s,List.assoc s bv),bv 
          with _ -> if List.assoc s rk = Equal 1 then (s,s),((s,s)::bv) else assert false in
        let rr,bbv = try (r,List.assoc r bbv),bbv 
          with _ -> if List.assoc r rk = Equal 1 then (r,r),((r,r)::bv) else assert false in
        GMsg (n,(ss,rr,(List.map (function ((m,lv),r) ->
                                    (m, lv, muelimination bbv None gend r)) ml)))
    | GASPoll(_,(v,r,g)) -> 
        let n = match next with None -> create_node_name () | Some k -> k in
        (* let k = create_node_name () in *)
        let cont = match gend with None -> GEnd (create_node_name ()) | Some g -> g in
        let bbv = (v,r)::bv in
        let lv = List.map fst (List.filter (function (vv,rr) -> r=rr) bv) in
        GPoll(n,((v,r),lv,muelimination bbv None None g,cont))
    | GASSeq (_,(g,r)) -> 
        (muelimination bv next (Some (muelimination bv None gend r)) g)
    | GASPar (_,(g,r)) -> 
        let n = match next with None -> create_node_name () | Some k -> k in
        let cont = match gend with None -> GEnd (create_node_name ()) | Some g -> g in
        GPar (n,(muelimination bv None None g ,muelimination bv None None r,cont))
  in
  match g with
      GASRec (_,g) -> 
        let zero = create_node_name () in
        GRec (zero,muelimination [] None (Some (GGoto 0)) g)
    | GASOne (_,g) -> 
        let zero = create_node_name () in
        let one = create_node_name () in
        GStart (zero,muelimination [] None (Some (GStop one)) g)
    | GASInv (_,g) -> 
        let zero = create_node_name () in
        let one = create_node_name () in
        GInv (zero,muelimination [] None (Some (GStop one)) g)
    | _ -> assert false

let linearity (g:global_type) =
  let rec get_sub lss pl r i =
    if List.mem (r^(string_of_int i)) lss || List.mem (r^(string_of_int i)) pl
    then get_sub lss pl r (i+1) 
    else (r^(string_of_int i),i) in
  let rec dequant ls = function 
    | GEnd n -> GEnd n 
    | GGoto n -> GGoto n 
    | GMsg (n,((vs,s),(vr,r),ml)) -> 
        let s = alias ls vs,s in 
        let r = alias ls vr,r in
        GMsg (n,(s,r,(List.map (function (m,lv,r) ->
                                  let llv = List.map (alias ls) lv in
                                  (m,llv,dequant ls r)) ml)))
    | GPoll (n,((x,r),pl,g,c)) -> 
        let pll = List.map (alias ls) pl in
        let lss = List.map snd ls in
        let s1,i = get_sub lss pll r 1 in
        let s2,_ = get_sub lss pll r (i+1) in
        let ls1 = (x,s1)::ls in
        let ls2 = (x,s2)::ls in
        GPar (n,(dequant ls1 g,dequant ls2 g,dequant ls c))
    | GPar (n,(g,r,c)) -> GPar (n,(dequant ls g,dequant ls r,dequant ls c))
    | GRec (n,g) -> GRec (n,dequant ls g) 
    | GStart (n,g) -> GStart(n,dequant ls g)
    | GInv (n,g) -> GInv(n,dequant ls g)
    | GStop n -> GStop n
  in
  let dqg = dequant [] g in
  let () = debug_wf ("Dequantification\n"^sprint_gt dqg) in
  let rec get_msg = function
    | GEnd n -> []
    | GGoto n -> []
    | GMsg (n,(s,r,ml)) -> 
        List.concat (List.map (function ((l,t),lv,c) ->
                                 (((s,r),l),(lv,t,c))::(get_msg c)) ml)
    | GPoll (n,((x,r),pl,g,c)) -> assert false
    | GPar (n,(g,r,c)) -> (get_msg g)@(get_msg r)@(get_msg c)
    | GRec (n,g) -> get_msg g
    | GStart (n,g) -> get_msg g
    | GInv (n,g) -> get_msg g
    | GStop n -> []
  in
  let lm = get_msg dqg in
  let () = 
     debug_wf ("List of messages\n"^ String.concat "\n" 
     (List.map (function (((s,r),l),(lv,t,c)) ->
     Printf.sprintf "%s>>%s:%s->%s:%s   {%s}(%s) ==> %s"
     l (fst s) (snd s) (fst r) (snd r)
     (String.concat "," lv) t (sprint_gt c)) lm)) in
  let rec compat lm = function
      [] -> true
    | (k,(lv,t,c))::q -> if List.mem_assoc k lm 
      then (
        let llv,tt,cc = List.assoc k lm in
        (lv <> llv || (
           let () = debug_wf ("Linearity WARNING: ambiguous message label "^snd k) in
           t=tt && c=cc)) && compat lm q
      )
      else compat lm q in
  let rec check = function
    | GEnd n -> true
    | GStop n -> true
    | GGoto n -> true
    | GMsg (n,(s,r,ml)) -> 
        List.for_all (fun x->x) (List.map (function (_,_,c) -> check c) ml)
    | GPoll (n,((x,r),pl,g,c)) -> assert false
    | GPar (n,(g,r,c)) -> 
        (check g) && (check r) && (check c) && (
          let lg = (get_msg g) in
          let lr = (get_msg r) in 
          compat lg lr)
    | GRec (n,g) -> check g
    | GStart (n,g) -> check g
    | GInv (n,g) -> check g
  in
  let () = debug_wf ("Linearity "^(string_of_bool (check dqg))) in 
  ()



let sanitize (rl:roles_list) (g:as_global)= 
  (* if it's a looping global type *)
  let g = global_conversion rl g in
  (* let rk = List.map (function (r,k,_)-> (r,k)) rl in *)
  let () = debug_wf ("Conversion from AST\n"^sprint_gt g) in
  let () = linearity g in
  g



let torole role g =
  debug_wf (Printf.sprintf "Projecting for %s" role);
  let nref = ref 0 in
  let create_node_name ()= let i = !nref in let _ = incr nref in i in
  let ndb = ref [] in
  let add_node n =
    if List.mem_assoc n !ndb
    then (List.assoc n !ndb,n)
    else 
      let i = create_node_name () in
      let () = ndb := (n,i)::!ndb in
      (i,n)
  in
  let next (i,n) =
    let ii = create_node_name () in
    let () = ndb := (n,ii)::!ndb in
    (ii,n) in
  let rec lmmerge lm = function
    | [] -> lm
    | ((l,t),lv,c)::q -> 
        if List.exists (function ((ll,_),lvv,cc)->ll=l && lv=lvv) lm
        then 
          let newlm = List.map (function ((ll,tt),lvv,cc) -> 
                                  if ll=l && lv=lvv
                                  then (if t=tt
                                        then ((ll,tt),lvv,merge [cc;c])
                                        else failwith "Unmergeable (different payload types)")
                                  else ((ll,tt),lvv,cc)) lm in
          lmmerge newlm q
        else
          lmmerge (((l,t),lv,c)::lm) q
  and ltequal = function
    | LEnd _ ,LEnd _ -> true
    | LSend (n,(r,ml)),LSend (m,(rr,mll)) -> 
        (r = rr && (try List.for_all (function ((l,lv,g),(ll,lvv,gg)) ->
                                        l = ll && lv = lvv && ltequal (g,gg))
                      (List.combine ml mll)
                    with Invalid_argument _ -> 
                      failwith "Unmergeable (different send operations)"))
    | LRecv (n,(s,ml)),LRecv (m,(ss,mll)) -> 
        (s = ss && (try List.for_all (function ((l,lv,g),(ll,lvv,gg)) ->
                                        l = ll && lv = lvv && ltequal (g,gg))
                      (List.combine ml mll)
                    with Invalid_argument _ -> false))
    | LPoll (n,(v,pl,g,c)),LPoll (m,(vv,pll,gg,cc))-> 
        (v = vv && pl = pll && ltequal (g,gg) && ltequal (c,cc))
    | LPar (n,(g,r,c)),LPar (m,(gg,rr,cc)) -> 
        ltequal (g,gg) && ltequal (r,rr) && ltequal (c,cc)
    | LRec (n,g),LRec (m,gg) -> ltequal (g,gg)
    | LStart (n,g),LStart (m,gg) -> ltequal (g,gg)
    | LInv (n,g),LInv (m,gg) -> ltequal (g,gg)
    | LStop _,LStop _ -> true
    | LGoto n,LGoto m -> n=m
    | LMu (n,g),gg -> ltequal (g,gg)
    | gg,LMu(n,g) -> ltequal (gg,g)
    | _ -> false
  and fusion = function 
      (*      | LEnd _ , c -> c *)
    | LStop (0,0) , c -> c
        (*      | c, LEnd _ -> c *)
    | c, LStop (0,0) -> c
    | LRecv(k,(s,lm)),LRecv (m,(ss,mll)) ->
        if s = ss then LMu(m,LRecv(k,(s,lmmerge lm mll)))
        else failwith "Unmergeable (different senders)"
    | LMu(k,g),c -> LMu(k,fusion (g,c))
    | g,LMu(k,c) -> LMu(k,fusion (g,c))
    | lt1,lt2 -> failwith ("Incompatible local types \n"^(sprint_lt lt1)
                           ^"\n"^(sprint_lt lt2))
  and merge2 l = match l with
      [] -> assert false
    | [lt] -> lt
    | c::d::q -> if ltequal (c,d) then merge2 (c::q)
      else let e = fusion (c,d) in merge2 (e::q)
                                     
  and merge l = match l with
      [] -> assert false
    | [lt] -> lt
    | c::q -> 
        if List.for_all (fun l -> ltequal (l,c)) q 
        then c 
        else match c,List.for_all (function LRecv(_) -> true|_->false) q with
          | LRecv(k,(s,lm)),true -> add (k,(s,lm)) q
          | _,_ -> failwith "Unmergeable (not only reception)"
  and add (k,(s,lm)) = function 
    | [] -> LRecv(k,(s,lm))
    | (LRecv(kk,(ss,lmm)))::q -> 
        if s = ss then
          let newlm = lmmerge lm lmm in
          add (k,(s,newlm)) q
        else failwith "Unmergeable (different senders)"
    | _ -> assert false
  in
  let rec aux mnode ls = function
    | GEnd n -> let k = add_node n in LEnd k
    | GStop n -> let k = add_node n in LStop k
    | GGoto n -> let k = add_node n in 
      if List.mem k mnode 
      then LStop (0,0) 
      else LGoto k
    | GMsg (n,(s,r,ml)) -> 
        let k = add_node n in 
        let kk = next k in
        let ss = alias ls (fst s),snd s in 
        let rr = alias ls (fst r),snd r in
        if ss = (role,role)
        then LSend(kk,(rr,
                       (List.map 
                          (function((l,t),lv,c)->
                             (l,t),(List.map (alias ls) lv),aux [] ls c
                          ) ml)))
        else if rr = (role,role)
        then LRecv(kk,(ss,
                       (List.map 
                          (function((l,t),lv,c)->
                             (l,t),(List.map (alias ls) lv),aux [] ls c
                          ) ml)))
        else 
          (debug_wf (Printf.sprintf "Introduction of Mu %d,%d " (fst k) (snd k)) ;
           LMu(kk,merge2 (List.map (function (_,_,c)-> aux (kk::mnode) ls c) ml)))
    | GPoll (n,((x,r),pl,g,c)) -> 
        let k = add_node n in 
        let kk = next k in 
        let kkk = next kk in 
        if r <> role then
          LPoll(k,((x,r),pl,aux [] ls g,aux [] ls c))
        else if List.mem role (List.map (alias ls) pl) then
          LPoll(k,((x,r),(List.map (alias ls) pl),aux [] ls g,aux [] ls c))
        else 
          LPar(k,(aux [] ((x,role)::ls) g,
                  LPoll(kk,((x,r),role::pl,aux [] ls g,LEnd kkk)),aux []ls c))
    | GPar (n,(g,r,c)) -> 
        let k = add_node n in 
        LPar(k,(aux [] ls g,aux [] ls r,aux [] ls c))
    | GRec (n,g) -> 
        let k = add_node n in LRec (k,aux [k] ls g)
    | GStart (n,g) -> 
        let k = add_node n in LStart (k,aux [] ls g)
    | GInv (n,g) -> 
        let k = add_node n in LInv (k,aux [] ls g)
                                
  in aux [] [] g

let project (rl:roles_list) (g:global_type) =
  let ltypes = List.map (function (role,_,_) -> (role,torole role g)) rl in
  let () = 
    List.iter (function (r,l) -> 
                 debug_wf ("Local type for "^r^":\n"^sprint_lt l)) ltypes in
  ltypes


(* Cleans the local types for empty parallel branches *)
let clean_local_role lt = 
  let rec lappend cc = function
    | LEnd n -> cc
    | LGoto n -> LGoto n 
    | LSend (n,((r,rr),ml)) ->
        let lmm = List.map (function (l,lv,g)-> (l,lv,lappend cc g)) ml in 
        LSend (n,((r,rr),lmm))
    | LRecv (n,((s,ss),ml)) -> 
        let lmm = List.map (function (l,lv,g)-> (l,lv,lappend cc g)) ml in 
        LRecv (n,((s,ss),lmm))
    | LPoll (n,((x,r),pl,g,c)) -> LPoll (n,((x,r),pl,g,lappend cc c))
    | LPar (n,(g,r,c)) -> LPar (n,(g,r,lappend cc c)) 
    | LRec (n,g) -> LRec (n,g)  
    | LMu (n,g) -> LMu (n,lappend cc g)
    | LStart (n,g) -> LStart (n,lappend cc g)
    | LInv (n,g) -> LInv (n,lappend cc g)
    | LStop n -> (match cc with LStop n -> LStop n | _ -> lappend (LStop n) cc)
  in
  let subst = ref [] in
  let rec aux = function 
    | LEnd n -> LEnd n 
    | LStop n -> LStop n 
    | LGoto n -> LGoto (re_alias !subst n)
    | LSend (n,((r,rr),ml)) ->
        let mll = List.map (function (l,lv,g)-> (l,lv,aux g)) ml in 
        LSend (n,((r,rr),mll))
    | LRecv (n,((s,ss),ml)) -> 
        let mll = List.map (function (l,lv,g)-> (l,lv,aux g)) ml in 
        LRecv (n,((s,ss),mll))
    | LPoll (n,((x,r),pl,LEnd _,c)) -> aux c
    | LPoll (n,((x,r),pl,LStop _,c)) -> aux c
    | LPoll (n,((x,r),pl,g,c)) -> LPoll (n,((x,r),pl,aux g,aux c))
    | LPar (n,(LEnd _,LEnd _,c)) ->  aux c
    | LPar (n,(LStop _,LStop _,c)) ->  aux c
    | LPar (n,(LEnd _,r,c)) -> lappend (aux c) (aux r)
    | LPar (n,(LStop _,r,c)) -> lappend (aux c) (aux r)
    | LPar (n,(g,LEnd _,c)) -> lappend (aux c) (aux g)
    | LPar (n,(g,LStop _,c)) -> lappend (aux c) (aux g)
    | LPar (n,(g,r,c)) -> LPar (n,(aux g,aux r,aux c)) 
    | LRec (n,g) -> LRec (n,aux g) 
    | LMu (n,g) -> let m = lnode_of_local g in (
        debug_wf (Printf.sprintf "Added subst %d,%d -> %d,%d" (fst n) (snd n) (fst m) (snd m));
        subst:=(n,m)::!subst; aux g)
    | LStart (n,g) -> LStart (n,aux g) 
    | LInv (n,g) -> LInv (n,aux g)
  in
  let rec apply x =
    let () = subst:= [] in
    let xx = aux x in
    if x = xx then x else apply xx 
  in
  apply lt
  

let clean_local ltl =
  let ltypes =  List.map (function (r,t)->(r,clean_local_role t)) ltl in
  let () = 
    List.iter (function (r,l) -> 
                 debug_wf ("Cleaned type for "^r^":\n"^sprint_lt l)) ltypes
  in
  ltypes
  
